#!/usr/bin/env python3
# coding: utf-8

from __future__ import division, print_function

import datetime
import re
import sys
import time

import six

from joker.cast import want_unicode


def seconds_to_hms(seconds):
    """
    >>> seconds_to_hms(4000)
    (1, 6, 40)
    :return:
    """
    # https://stackoverflow.com/a/775075/2925169
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    return h, m, s


def eazy_date(x):
    """  
    >>> eazy_date(0)
    datetime.date(2017, 5, 5) 
    >>> eazy_date('today')
    datetime.date(2017, 5, 5) 
    
    >>> eazy_date(-1)
    datetime.date(2017, 5, 4) 
    >>> eazy_date('yesterday')
    datetime.date(2017, 5, 4) 
    
    >>> eazy_date(1)
    datetime.date(2017, 5, 6) 
    >>> eazy_date('tomorrow')
    datetime.date(2017, 5, 6) 
    
    >>> eazy_date(datetime.date.today())
    datetime.date(2017, 5, 5) 
    >>> eazy_date(datetime.datetime.now())
    datetime.date(2017, 5, 5) 
    
    >>> eazy_date('20170505')
    datetime.date(2017, 5, 5) 
    >>> eazy_date('2017-05-06')
    datetime.date(2017, 5, 6) 
    >>> eazy_date('05-06')
    datetime.date(2017, 5, 6) 
    >>> eazy_date('0506')
    datetime.date(2017, 5, 6) 
    
    :param x: 
    :return: a datetime.date instance
    """
    day = datetime.timedelta(days=1)
    today = datetime.date.today()
    if isinstance(x, int):
        return today + x * day
    elif isinstance(x, datetime.date):
        return x
    elif isinstance(x, datetime.datetime):
        return x.date()
    elif isinstance(x, six.string_types):
        x = want_unicode(x).lower()
    else:
        t = x.__class__.__name__
        raise TypeError('cannot convert type {} to date'.format(t))

    if x == 'today':
        return eazy_date(0)
    if x == 'yesterday':
        return eazy_date(-1)
    if x == 'tomorrow':
        return eazy_date(1)

    if re.match(r'\d{8}$', x):
        return datetime.datetime.strptime(x, '%Y%m%d').date()

    if re.match(r'\d{4}$', x):
        x = '{}{}'.format(today.year, x)
        return datetime.datetime.strptime(x, '%Y%m%d').date()

    if re.match(r'\d{4}-\d{1,2}-\d{1,2}$', x):
        return datetime.datetime.strptime(x, '%Y-%m-%d').date()

    if re.match(r'\d{1,2}-\d{1,2}$', x):
        x = '{}-{}'.format(today.year, x)
        return datetime.datetime.strptime(x, '%Y-%m-%d').date()
    raise ValueError('unknow date format')


def date_range(start, stop=0, step=1):
    """
    >>> list(date_range(-3, 0))  # last 3 days
    [datetime.date(2017, 5, 2),
     datetime.date(2017, 5, 3),
     datetime.date(2017, 5, 4)]
     
    :param start: 
    :param stop: 
    :param step: 
    :return: 
    """
    start = eazy_date(start)
    stop = eazy_date(stop)
    delta = datetime.timedelta(days=step)
    while (start - stop).total_seconds() * step < 0:
        yield start
        start += delta


class Timer(object):
    def __init__(self, name=''):
        self.name = name

    def __enter__(self):
        self.time = time.time()
        return self

    def __exit__(self, typ, value, traceback):
        interval = time.time() - self.time
        p = 'Timer {}'.format(self.name) if self.name else 'Timer'
        msg = '{}: {} sec'.format(p, interval)
        print(msg, file=sys.stderr)


class TimeMachine(object):
    def __init__(self, start=None, speed=None):
        self._initpoint = datetime.datetime.now()
        self._imaginary = not (start or speed)
        self._start = start or self._initpoint
        self._speed = speed or 1.

    def now(self):
        if not self._imaginary:
            return datetime.datetime.now()
        delta = datetime.datetime.now() - self._initpoint
        return self._start + delta * self._speed

    @staticmethod
    def convert_time_to_timedelta(t):
        return datetime.timedelta(
            hours=t.hour, minutes=t.minute, seconds=t.second,
            microseconds=t.microsecond
        )


class TimeSlicer(TimeMachine):
    EPOCH = datetime.datetime(1970, 1, 1)

    def __init__(self, start=None, speed=None, ts_size=600):
        super(TimeSlicer, self).__init__(start, speed)
        self._ts_delta = datetime.timedelta(seconds=ts_size)

    def get_current_timeslice(self, relative=True):
        dt = self.now()
        return self.convert_datetime_to_timeslice(dt, relative=relative)

    @staticmethod
    def guess_slice_size(ts):
        """
        assuming slice size a multiple of 60
        :param ts:
        :return:
        """
        t = time.time()
        return int(round(t / ts / 60.)) * 60

    def convert_timeslice_to_time(self, ts):
        dt = self.EPOCH + self._ts_delta * ts
        return dt.time()

    def convert_timeslice_to_datetime(self, ts, relative=True):
        """
        If relative is true, given ts is considered relative to
        00:00:00 of today.

        :param ts: an integer, serial number of timeslice
        :param relative: bool
        :return: a datetime instance
        """
        if not relative:
            return self.EPOCH + self._ts_delta * ts
        d = self.now().date()
        t = self.convert_timeslice_to_time(ts)
        return datetime.datetime.combine(d, t)

    def convert_time_to_timeslice(self, t):
        delta = self.convert_time_to_timedelta(t)
        # use .total_seconds to be compat with python 2.x
        return int(delta.total_seconds() / self._ts_delta.total_seconds())

    def convert_datetime_to_timeslice(self, dt, relative=True):
        if relative:
            return self.convert_time_to_timeslice(dt.time())
        delta = dt - self.EPOCH
        # use .total_seconds to be compat with python 2.x
        return int(delta.total_seconds() / self._ts_delta.total_seconds())




#!/usr/bin/env python3
# coding: utf-8

from __future__ import division, print_function

import datetime
import json


class JSONEncoderExtended(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, 'as_json_serializable'):
            return obj.as_json_serializable()
        datetime_types = datetime.datetime, datetime.time, datetime.time
        if isinstance(obj, datetime_types):
            return obj.isoformat()
        elif isinstance(obj, datetime.timedelta):
            return obj.total_seconds()
        else:
            return self.default(vars(obj))


def human_json_dumps(s, **kwargs):
    return json.dumps(s, indent=4, ensure_ascii=False,
                      cls=JSONEncoderExtended, **kwargs)



#!/usr/bin/env python3
# coding: utf-8

from __future__ import division, unicode_literals

from copy import deepcopy

from joker.cast.iterative import reusable


def get_parent_from_ascendants(ascendants):
    return ([0] + ascendants)[-1]


def compact_tree_assemble(records):
    """
    flat -> nested
    :param records: a list of dicts
    :return: (tree, tmap), both are dicts
    """
    records = reusable(records)
    tmap = {rec['id']: dict() for rec in records}
    tree = {}
    for rec in records:
        i = rec['id']
        p = rec['parent']
        pnode = tmap.get(p, tree)
        inode = tmap[i]
        pnode[i] = inode
    return tree, tmap


def standard_tree_assemble(records, setnone=False):
    """
    flat -> nested
    :param records: a list of dicts
    :param setnone: bool, default False
    :return: (tree, tmap), both are dicts
    """
    records = reusable(records)
    ids = [rec['id'] for rec in records]
    tmap = {i: {'id': i, 'children': list()} for i in ids}
    tree = {'id': 0, 'children': list()}
    for rec in records:
        i = rec['id']
        p = rec['parent']
        pnode = tmap.get(p, tree)
        inode = tmap[i]
        pnode['children'].append(inode)
    if setnone:
        for val in tmap.values():
            if not val['children']:
                val['children'] = None
    return tree, tmap


def standard_tree_dissemble(tree):
    """
    nested structure -> flat records
    :param tree: a nested dict
    :return: a list of dicts
    """
    tree = deepcopy(tree)
    tree['ascendants'] = []
    stack = [tree]
    collect = []
    while stack:
        node = stack.pop()
        collect.append(node)
        ascendants = node['ascendants'] + [node['id']]
        for child in node.get('children', []):
            child['ascendants'] = ascendants
            stack.append(child)
    for node in collect:
        try:
            del node['children']
        except KeyError:
            pass
    return collect


def compact_tree_dissemble(tree):
    """
    nested -> flat
    :param tree: a nested dict
    :return: a list of dicts
    """
    tree = deepcopy(tree)
    # tree['ascendants'] = []
    stack = [(i, list(), node) for (i, node) in tree.items()]
    collect = list()
    while stack:
        i, ascendants, node = stack.pop()
        collect.append({'id': i, 'ascendants': ascendants})
        ascendants = ascendants + [i]
        if node is None:
            continue
        for k, child in node.items():
            stack.append((k, ascendants, child))
    return collect



#!/usr/bin/env python3
# coding: utf-8

from __future__ import division, print_function

import os
import re
import sys


def under_package_dir(package, *paths):
    if isinstance(package, str):
        import importlib
        package = importlib.import_module(package)
    p_dir = os.path.dirname(package.__file__)
    return os.path.join(p_dir, *paths)


def load_modules_under_package(package, module_names=None):
    import importlib
    if isinstance(package, str):
        package = importlib.import_module(package)
    prefix = package.__name__
    match = re.compile(r'^[^_].*\.pyc?$').match

    # by default import all modules under package
    if module_names is None:
        d = os.path.split(package.__file__)[0]
        module_names = (x for x in os.listdir(d) if match(x))
        module_names = (x.split('.')[0] for x in module_names)
        module_names = list(module_names)

    loaded = list()
    for name in module_names:
        mp = '{}.{}'.format(prefix, name)
        loaded.append(importlib.import_module(mp))
    return loaded


def under_home_dir(*paths):
    if sys.platform == 'win32':
        homedir = os.environ["HOMEPATH"]
    else:
        homedir = os.path.expanduser('~')
    return os.path.join(homedir, *paths)


def under_joker_dir(*paths):
    p = under_home_dir('.joker')
    p = os.environ.get('JOKER_HOMEDIR', p)
    p = os.path.expanduser(p)
    p = os.path.abspath(p)
    return os.path.join(p, *paths)


def make_joker_dir():
    # silently return if userdir_path exists as a dir
    d = under_joker_dir()
    if not os.path.isdir(d):
        os.mkdir(d, int('700', 8))


def keep_file_extension(old_path, new_path):
    _, old_ext = os.path.splitext(old_path)
    p, new_ext = os.path.splitext(new_path)
    if old_ext.lower() == new_ext.lower():
        return new_path
    return os.path.join(p, old_ext)


def url_to_filename(url):
    # http://stackoverflow.com/questions/295135/
    name = re.sub(r'[^\w\s_.-]+', '-', url)
    return re.sub(r'^{http|https|ftp}', '', name)


def validate_ipv4_address(address):
    import socket
    # http://stackoverflow.com/a/4017219/2925169
    try:
        socket.inet_pton(socket.AF_INET, address)
    except AttributeError:  # no inet_pton here, sorry
        try:
            socket.inet_aton(address)
        except socket.error:
            return False
        return address.count('.') == 3
    except socket.error:  # not a valid address
        return False
    return True


def validate_ipv6_address(address):
    import socket
    # http://stackoverflow.com/a/4017219/2925169
    try:
        socket.inet_pton(socket.AF_INET6, address)
    except socket.error:  # not a valid address
        return False
    return True



#!/usr/bin/env python3
# coding: utf-8

from __future__ import division, print_function

import codecs
import collections
import json

import six

__version__ = '0.0.17'


def regular_cast(original, *attempts):
    for attem in attempts:
        if not callable(attem):
            return attem
        try:
            return attem(original)
        except (TypeError, ValueError):
            pass
    return original


def smart_cast(value, default):
    """
    Cast to the same type as `default`;
    if fail, return default
    :param value:
    :param default:
    :return:
    """
    func = type(default)
    try:
        return func(value)
    except (TypeError, ValueError):
        return default


def numerify(s):
    return regular_cast(s, int, float)


def want_bytes(s, **kwargs):
    """
    :param s: 
    :param kwargs: key word arguments passed to str.encode(..)
    :return: 
    """
    if not isinstance(s, six.binary_type):
        s = s.encode(**kwargs)
    return s


def want_unicode(s, **kwargs):
    """
    :param s: 
    :param kwargs: key word arguments passed to bytes.decode(..)
    :return: 
    """
    if not isinstance(s, six.text_type):
        return s.decode(**kwargs)
    return s


def want_str(s, **kwargs):
    """
    :param s:
    :param kwargs: key word arguments passed to s.decode(..)
    :return:
    """
    if not isinstance(s, str):
        return s.decode(**kwargs)
    return s


def namedtuple_to_dict(nt):
    fields = getattr(nt, '_fields')
    return collections.OrderedDict(zip(fields, nt))


def represent(obj, params):
    """
    :param obj:
    :param params: a dict or list
    """
    c = obj.__class__.__name__
    if isinstance(params, dict):
        parts = ('{}={}'.format(k, repr(v)) for k, v in params.items())
    else:
        parts = ('{}={}'.format(k, repr(getattr(obj, k))) for k in params)
    s = ', '.join(parts)
    return '<{}({}) at {}>'.format(c, s, hex(id(obj)))


def indented_json_print(o, *args, **kwargs):
    # https://stackoverflow.com/a/12888081/2925169
    decode = codecs.getdecoder('unicode_escape')
    outstr = json.dumps(o, indent=4, sort_keys=True)
    print(decode(outstr)[0], *args, **kwargs)



#!/usr/bin/env python3
# coding: utf-8

from __future__ import division, print_function

import functools
import inspect
import itertools


def noop(*_, **__):
    pass


def adaptive_call(entry):
    """
    >>> import sys
    >>> entry = [print, ['a', 'b'], {'file': sys.stderr}]
    >>> adaptive_call(entry)

    :param entry: an iterable or a callable
    :return:
    """
    pargs = []
    kwargs = dict()
    if callable(entry):
        return entry()

    items = list(entry)
    if not items:
        return items
    if not callable(items[0]):
        raise TypeError('first item of entry must be a callable')

    for x in items[1:]:
        if isinstance(x, (list, tuple)):
            pargs.extend(x)
        elif isinstance(x, dict):
            kwargs.update(x)
        else:
            raise TypeError('params must be a tuple, list or dict')
    return items[0](*pargs, **kwargs)


def format_class_path(obj):
    if isinstance(obj, type):
        klass = obj
    else:
        klass = type(obj)
    m = getattr(klass, '__module__', None)
    q = getattr(klass, '__qualname__', None)
    n = getattr(klass, '__name__', None)
    name = q or n or ''
    if m:
        return '{}.{}'.format(m, name)
    return name


def format_function_path(func):
    if not inspect.ismethod(func):
        mod = getattr(func, '__module__', None)
        if mod is None:
            return func.__qualname__
        else:
            return '{}.{}'.format(mod, func.__qualname__)
    klass_path = format_class_path(func.__self__)
    return '{}.{}'.format(klass_path, func.__name__)


# compatibility
fmt_class_path = format_class_path
fmt_function_path = format_function_path


def instanciate(cls):
    return cls()


def instanciate_with_foolproof(cls):
    """
    The return class can be called again without error
    """
    if '__call__' not in cls.__dict__:
        cls.__call__ = lambda x: x
    return cls()


class AttrEchoer(object):
    """
    Resembles an enum type
    Reduces typos by using syntax based completion of dev tools
    
    Example:
        
        @instanciate_with_foolproof
        class Event(AttrEchoer):
            _prefix = 'event'
            bad_params = ''  # assign whatever
            unauthorized_access = ''  
            undefined_fault = ''
            ...
       
        # no error: 
        assert Event.unauthoried  == 'event.bad_params'
    """
    _prefix = '_root.'

    def __init__(self):
        pass

    def __getattribute__(self, key):
        kls = type(self)
        if key in kls.__dict__ and key != '_prefix':
            if not kls._prefix:
                return key
            return '{}{}'.format(kls._prefix, key)
        return object.__getattribute__(self, key)


def multilevel_get(d, *keys):
    for k in keys:
        v = d.get(k)
        if v is None:
            return None
        d = v
    return d


def _first_arg(*a):
    return a[0]


class ConstantCallable(object):
    def __init__(self, value):
        self.value = value

    def __call__(self, *args, **kwargs):
        return self.value


_always_return_true = ConstantCallable(True)
_always_return_false = ConstantCallable(False)


class Void(object):
    """
    Act as 0, False, '', [] 
    """
    __bool__ = _always_return_false
    __nonzero__ = _always_return_false
    __add__ = _first_arg
    __sub__ = _first_arg
    __radd__ = _first_arg
    __rsub__ = _first_arg
    __round__ = _first_arg
    __truediv__ = _first_arg
    __floordiv__ = _first_arg
    __rtruediv__ = _first_arg
    __rfloordiv__ = _first_arg
    __rmul__ = _first_arg
    __gt__ = _always_return_false
    __ge__ = _always_return_false
    __lt__ = _always_return_false
    __le__ = _always_return_false
    __len__ = ConstantCallable(0)

    def __init__(self, symbol='-'):
        self.symbol = symbol

    def __str__(self):
        return self.symbol

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, repr(self.symbol))

    def __eq__(self, other):
        if isinstance(other, Void):
            return True
        return False


class Universe(object):
    __contains__ = _always_return_true
    __iter__ = ConstantCallable(tuple())


# TODO: suport negative index (castfunc=-1 to get last item)
def castable(func):
    """
    >>> @castable
    ... def myfunc(*args):
    ...     for i in range(*args):
    ...         yield i
    ...
    >>> myfunc(12, castfunc=tuple)
    (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11)
    >>> myfunc(0, 12, 2, castfunc=2)
    (0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11)
   
    Purely syntax sugar,
    to make interactive use of some functions easier.
    Cast a generator function to list, set, or select n-th item, etc.

        myfunc(..., castfunc=list)   <=>  list(myfunc(...))
        myfunc(..., castfunc=1)      <=>  list(myfunc(...))[1]
    """
    @functools.wraps(func)
    def _decorated_func(*args, **kwargs):
        castfunc = None
        if 'castfunc' in kwargs:
            castfunc = kwargs['castfunc']
            del kwargs['castfunc']

            # shortcut to pick up nth record
            if isinstance(castfunc, int):
                n = castfunc
                castfunc = lambda r: next(itertools.islice(r, n, None))

        result = func(*args, **kwargs)
        if castfunc:
            result = castfunc(result)
        return result
    return _decorated_func



#!/usr/bin/env python3
# coding: utf-8

from __future__ import division, print_function

import collections
import sys

import six


class Traverser(object):
    handlers = dict()

    def __init__(self, obj, consumer):
        self.queue = collections.deque()
        self.queue.append(obj)
        self.consumer = consumer

    def traverse(self):
        kls = self.__class__
        while self.queue:
            obj = self.queue.popleft()
            for typ in obj.__class__.mro():
                if typ in self.handlers:
                    kls.handlers[typ](self, obj)
                    break


def register_handler(*types):
    def decorator(func):
        for typ in types:
            Traverser.handlers[typ] = func
        return func
    return decorator


@register_handler(int, float)
def handle_primatives(traverser, obj):
    s = '<{}:{}>'.format(obj.__class__.__name__, obj)
    traverser.consumer(s)


@register_handler(*six.string_types)
def handle_primatives(traverser, obj):
    if isinstance(obj, six.binary_type) and sys.version_info.major == 2:
        obj = obj.decode()
    s = '<str:{}>'.format(obj)
    traverser.consumer(s)


@register_handler(set)
def handle_set(traverser, obj):
    traverser.consumer('<set>')
    items = list(obj)
    items.sort()
    for item in items:
        traverser.queue.append(item)


@register_handler(list)
def handle_list(traverser, obj):
    traverser.consumer('<list>')
    for item in obj:
        traverser.queue.append(item)


@register_handler(dict)
def handle_dict(traverser, obj):
    traverser.consumer('<dict>')
    keys = list(obj)
    keys.sort()
    for k in keys:
        traverser.queue.append(k)
        traverser.queue.append(obj[k])


# the last resort
@register_handler(object)
def handle_object(traverser, _):
    traverser.consumer('<object>')



#!/usr/bin/env python3
# coding: utf-8

from __future__ import division, print_function

import collections


class DefaultOrderedDict(collections.OrderedDict):
    # Source: http://stackoverflow.com/a/6190500/562769
    def __init__(self, default_factory=None, *a, **kw):
        if (default_factory is not None and
                not isinstance(default_factory, collections.Callable)):
            raise TypeError('first argument must be a callable')
        collections.OrderedDict.__init__(self, *a, **kw)
        self.default_factory = default_factory

    def __getitem__(self, key):
        try:
            return collections.OrderedDict.__getitem__(self, key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = value = self.default_factory()
        return value

    def __reduce__(self):
        if self.default_factory is None:
            args = tuple()
        else:
            args = self.default_factory,
        return type(self), args, None, None, self.items()


class RecursiveDefaultDict(collections.defaultdict):
    def __init__(self, **kwargs):
        super(RecursiveDefaultDict, self).__init__(self.__class__, **kwargs)


class RecursiveCounter(RecursiveDefaultDict):
    def total(self, as_string=False):
        s = 0
        for v in self.values():
            if isinstance(v, six.integer_types) or isinstance(v, float):
                s += v
            elif isinstance(v, RecursiveCounter):
                s += v.total()
        if as_string:
            s = str(s)
        return s
